# HPIPMIevents
"""Convert a record from a DBus GetBulk response to a more expressive object.
   SEL_entry - for SEL and FPL type E0/E1 entries
   IML_entry - for IML entries (future)
   Each class includes a Physloc instance"""

from __future__ import absolute_import

# import pdb
import logging
import sys
import time

from Engines.Indictment.Indictable import Indictable
from RASutils.HPphysloc import Physloc

sys.modules[__name__].HP = True

class SEL_entry(Indictable):
    """SEL and FPL type E0/E1 entries.  Type 02 entries are not handled."""

    def _indicted(self):
        """Called when an event has advanced to "Indicted' status."""
        logging.info("SEL indict CB: %s", repr(self.physloc))

    def __init__(self, hpipmidRecord = None):
        """Unpack a single record from a GetBulk call against hpipmid.
        When called with no argument, it generates a zero-filled record
        for testing."""

        # Inherited from Indictable
        Indictable.__init__(self)
        self.IndictedCB = self._indicted
        self.CPH = 3
        E = self.E

        if not hpipmidRecord:
            E.index, E.RecID, E.Timestamp, E.RecType, E.EventID, \
                E.RepEntID, E.ExRepEntID, E.AlertLevel, E.DataType = [ 0 ] * 9
            E.Data = [ 0 ] * 8
            E.Desc = ""
            self.physloc = Physloc()    # yes, on self
            return

        # Subclass extensions
        if (len(hpipmidRecord) != 13):
            raise TypeError
        E.index, E.RecID, E.Timestamp, E.RecType, \
        E.EventID, E.RepEntID, E.ExRepEntID, E.AlertLevel, \
        E.DataType, DataRaw, E.Desc, PhyslocRaw, \
        PhyslocDesc = hpipmidRecord
        if E.RecType != 0xE0 and E.RecType != 0xE1:
            raise TypeError
        E.EventID = int(E.EventID)    # convert from DBus.uint

        # DataRaw and PhyslocRaw should be dbus.ByteArray, a subtype
        # of str.  This was set in the interface object call.
        E.Data = [ ord(i) for i in DataRaw ]
        self.physloc = Physloc(PhyslocRaw, PhyslocDesc)
        self.IndexStr = "%s(%X)" % (self.__class__.__name__, E.RecID)

    def __hash__(self):
        """Concatenate index, RecID, and timestamp."""
        E = self.E
        return hash(((E.index & 0xFFFFFFFF) << 48) + \
                    ((E.RecID & 0xFFFFFFFF) << 32) + \
                      E.Timestamp)

    def __eq__(self, other):
        """Compare all fields."""
        E = self.E
        O = other.E
        return (E.index      == O.index) and \
               (E.RecID      == O.RecID) and \
               (E.Timestamp  == O.Timestamp) and \
               (E.RecType    == O.RecType) and \
               (E.EventID    == O.EventID) and \
               (E.RepEntID   == O.RepEntID) and \
               (E.ExRepEntID == O.ExRepEntID) and \
               (E.AlertLevel == O.AlertLevel) and \
               (E.DataType   == O.DataType) and \
               (self.physloc == other.physloc)

    def __str__(self):
        """Returns index, RecID, timestamp, Rectype, EventID."""
        E = self.E
        return "%1d:0x%04X @ %s type %02X ID %04d" % \
            (E.index, E.RecID, time.ctime(E.Timestamp), E.RecType, E.EventID)
        # s = s + "%s\n" % ( self.Desc )
        # s = s + "\tb9 = 0x%02X %02X %02X %02X %02X %02X %02X %02X\n" % \
        #     ( self.Data[0], self.Data[1], self.Data[2], self.Data[3],
        #       self.Data[4], self.Data[5], self.Data[6], self.Data[7] )
        # s = s + "%s\n" % ( self.physloc )

class IML_entry(Indictable):
    """ProLiant Integrated Management Log (IML) entries."""

    def _indicted(self):
        """Called when an event has advanced to "Indicted' status."""
        logging.info("IML indict CB: %s", repr(self.physloc))

    def __init__(self):
        """Unpack a single record from a GetBulk call against hpipmid."""

        # Inherited from Indictable
        Indictable.__init__(self)
        self.IndictedCB = self._indicted

        # Subclass extensions
        raise TypeError

if __name__ == "__main__":
    s1 = SEL_entry()
    s2 = SEL_entry([ 1, 2, 3, 4, 5, 6, 7, 8, 9,
        '\0\1\2\3\4\5\6\7', "Ex data bytes", '\0\1\2\3\4\5\6\7', "Physloc"])
