# IPMIlogs
"""Receive log updates from hpipmid and compare them against trend lists"""

from __future__ import absolute_import,  with_statement   # Not needed at 2.6

# import pdb
import logging
import os
import sys

from RASutils import EventManager as REM
from RASutils.DBus import HPDBus
from .HPIPMIevents import SEL_entry

sys.modules[__name__].HP = True

# This seems sloppy to leave these out here but I'll need them for
# IML some day...

EZBMC_LOG_FIRST_RECORD = 0
EZBMC_LOG_LAST_RECORD  = 0xFFFFFFFF

class SEL_snapline(object):
    """Parse one line from a SEL group file, which is a CEDB snapshot."""

    def __init__(self, oneline):
        """JFDI"""
        self.Type, \
        self.EventID, \
        self.Severity, \
        self.Summary, \
        self.Provider, \
        self.Indication = oneline.split('^')
        if str(self.EventID).find(':') == -1:        # Type E0 or E1
            self.EventID = int(self.EventID)
        else:
            self.EventID = -1                   # Not doing type 02 for now

    def __str__(self):
        """Really just for development."""
        return "%5d (%8s): %s" % (self.EventID, self.Severity, self.Summary)

class SEL_group(object):
    """Construct an object of SEL snapshots from a group/trend file."""

    All = [ ]           # Read in from files

    def __init__(self, groupname):
        """Read one file of trend triggers."""
        self.Name = groupname
        self.Events = [ ]
        with open(groupname, "r") as f:
            self.Events = [SEL_snapline(l) for l in f.readlines()]
        self.EventSet = frozenset([e.EventID for e in self.Events])

    def __str__(self):
        """Really just for development."""
        s = "%s %s" % ('-' * 10, self.Name)
        for e in self.Events:
            s += "\n" + str(e)
        return s

    @staticmethod
    def Setup():
        """Read the trend files."""
        # cwd is still "root" of the project
        oldbase = os.getcwd()
        newbase, junk = os.path.split(sys.modules[__name__].__file__)
        try:
            os.chdir(newbase + "/SELtrends")
            SEL_group.All = [ SEL_group(fname) for fname in os.listdir('.') if
                os.path.isfile(fname) ]
        except Exception:
            logging.exception("Can't get SEL error groups")
        finally:
            os.chdir(oldbase)
        return True

#############################################################################

class LogHandler(object):
    """Declare DBus 'Logs' interface hander for SEL and IML.  It's the
       gateway to indictment processing.
    """

    @staticmethod
    def _GetTailSpan(PathSpanDuple):
        """Retrieve the event span at the end of the log.  This is a timeout
        callback so return False to avoid rescheduling it."""
        base, span = PathSpanDuple
        DBusEvents, nextIndex, nextRecID = HPDBus.GetBulk(
            base,
            EZBMC_LOG_LAST_RECORD,      # cache index
            EZBMC_LOG_LAST_RECORD,      # IPMI record ID
            False,                      # True == forward, false == backward
            span)                       # How many
        logging.debug("%s GetBulk(%d)->%d, next I:R=%d:0x%08X",
            base, span, len(DBusEvents), nextIndex, nextRecID)

        # Real events came in with a negative span (seconds).  Test
        # events are with a positive number.
        if base == "SEL":
            current_events = [ SEL_entry(i) for i in DBusEvents ]
            if span > 0:        # Testing
                for c in current_events:
                    c.arrivalTS = int(c.E.Timestamp)
            for g in SEL_group.All:
                for c in current_events:
                    if c.E.EventID in g.EventSet:
                        c.Indict()

        return False    # Don't reschedule the timeout that got me here

    @staticmethod
    def _LogSigHandler(count, onpath = None, onmember = None):
        """Receive DBus signals from hpipmid 'Logs' interface and buffer
           them via a settling time delay to avoid retrieval floods."""

        span = 0
        junk, onbase = os.path.split(str(onpath))   # from dbus.ObjectPath
        logging.info("SIGNAL '%s' %d on %s", onmember, count, onpath)
        if count < 0:
            span = -3600	# Actual appends, grab this many seconds of events
        elif count > 0:
            span = count	# Testing, grab this many events
        else:
            return
        # Timeout is one longer than hpipmid short sampling
        REM.TimeoutAdd(5, LogHandler._GetTailSpan, (onbase, span))

    @staticmethod
    def Setup():
        """Set up the signal handlers for SEL and IML."""
        ret = False
        msg = "Couldn't add SEL signal receiver"
        try:
            ret = HPDBus.AddSigHandler(
                HPDBus.SELobjpath,
                HPDBus.LogsInterface,
                HPDBus.SigAppended,
                LogHandler._LogSigHandler)
            if not ret:
                logging.error(msg)
        except DBusException, e:
            msg = "%s: %s" % str(e)
            logging.exception(msg)
            ret = False
        return ret

#############################################################################

def Setup(fromMain=False):
    """Set up LogHandler, SEL, and IML classes."""
    LogHandler.Setup()
    SEL_group.Setup()

if __name__ == "__main__":
    Setup(True)
