"""Organize collections of troubled physlocs."""

from __future__ import absolute_import

# import pdb
import logging
import sys
from time import time as NOW

from RASconfig import Globals
from RASutils import EventManager as REM       # classless
from RASutils.DBus import HPDBus, EmitSignal   # two classes

sys.modules[__name__].HP = True

class Indicted(object):
    """Wrap a list of indicted events with an expiration timestamp."""
    def __init__(self, Ilist, deltaT):
        self.Ilist = Ilist
        self.expireTS = int(NOW()) + deltaT

    def alive(self, timeref):
        """Compare internal expiration to reference, returns Boolean."""
        return self.expireTS > int(timeref)

    def expired(self, timeref):
        """Compare internal expiration to reference, returns Boolean."""
        return self.expireTS < int(timeref)

class IndictQueues(object):
    """No instances should ever be created."""

    Incoming    = [ ]   # Raw Indictables from analyzers
    Candidates  = { }   # Indictables keyed by physloc ("spans")
    Indictments = { }   # Spans that exceed rate/burst limits

    IncomingTotal      = 0  # Count of AddIndictable() calls
    IncomingUnique     = 0
    IndictmentsExpired = 0

    # Seconds.  TTLsuppress is the time period that a duplicate error/
    # action should be held off, even if the device continues to complain.
    # TTLincoming is an assist to avoid bursts from analyzers
    # especially during testing.  It needs to be longer than any rate
    # calculation base, but half the larger TTLsuppress (Nyquist).
    TTLsuppress = 60 * 60 * 24
    TTLincoming = int(TTLsuppress/3) - 100

    # I've been assigned EventID 8338, "OS_RAS_CUM_RATE" == 0x2092.
    # REID class in CEDB is OS agent (17), just use REID = 0x00.
    # Alert level = 3, data type = 4 => 0x64
    # FIXME: make this a delegate to HPIPMIlogs, registered here from there
    @classmethod
    def SELappend(cls, PL):
        """Wrapper to emit a DBus signal with a Physloc argument."""
        logging.warning("'SEL.Append' for PL %r", PL)
        raw = "\xE0\x00\x92\x20\x80\x64" + PL.raw
        HPDBus.SEL_append_raw(raw)

    @staticmethod
    def _spanalyze(CandIs):
        lowCPH = 120
        for I in CandIs:
            if 1 < I.CPH < lowCPH:
                lowCPH = I.CPH
        if lowCPH > 0:
            burstSBE = int(360 / lowCPH)
            if burstSBE < 15:
                burstSBE = 15
        else:
            lowCPH = 4
            burstSBE = 90
        return (lowCPH, burstSBE)

    # For each physloc there can be multiple events, each with a unique
    # Count Per Hour (CPH).  Use the 'most sensitive' (smallest) CPH to
    # analyze the time span for that physloc's events.  Short bursts are
    # also evaluated in terms of Seconds Between Events (SBE).
    @classmethod
    def Candidate2Indicted(cls, cbdata=None):
        """Event loop idle callback, chained from FilterIncoming()."""

        zerolen = [ ]   # Accumulate changes to Candidates (because
        trimmed = { }   # it's rude to modify a loop iterator)
        for PL, CandIs in cls.Candidates.iteritems():
            if len(CandIs) < 2:          # No rates/trends with only 1 item
                if not len(CandIs):      # Idiot check
                    zerolen.append(PL)
                continue
            lowCPH, burstSBE = cls._spanalyze(CandIs)

            # As long as there are enough events, see if the span
            # CandIs[0:lowCPH] happened in < 1 hour.  Then see if that span
            # runs longer than lowCPH.  Extract that longest span.  In any
            # case, pop CandIs[0] and repeat.
            startlen = len(CandIs)
            while len(CandIs) >= lowCPH:
                last = lowCPH - 1
                if CandIs[last].TS - CandIs[0].TS > 3600:
                    CandIs.pop(0)
                    continue
                last += 1                   # How long does it last?
                while last < len(CandIs) and \
                        CandIs[last].TS - CandIs[0].TS < 3600:
                    last += 1
                span = CandIs[:last]
                if PL not in cls.Indictments:  # Start a new one
                    cls.Indictments[PL] = Indicted(span, cls.TTLsuppress)
                    cls.SELappend(PL)
                else:                       # Add them, but not dupes
                    Ilist = cls.Indictments[PL].Ilist
                    newIs = [I for I in span if I not in Ilist]
                    Ilist.extend(newIs)
                    span = newIs
                for I in span:       # notify/log new contributors
                    # FIXME: SQL modify these to "Indicted"
                    if callable(I.IndictedCB):
                        I.IndictedCB()
                CandIs.pop(0)               # and go around one more time

            if len(CandIs) != startlen:
                trimmed[PL] = CandIs

        # No longer rude to modify Candidates
        for PL, span in trimmed.iteritems():
            if len(span):
                cls.Candidates[PL] = span
            else:
                del cls.Candidates[PL]
        for PL in zerolen:
            del cls.Candidates[PL]
        logging.debug("Candidates -----------------------")
        for PL, Ilist in cls.Candidates.iteritems():
            logging.debug("%r: %s", PL, [I.IndexStr for I in Ilist])
        logging.info("Indictments ----------------------")
        for PL, IndictedList in cls.Indictments.iteritems():
            EmitSignal.dispatch("IndictedPhysloc", repr(PL))
            logging.info("%r: %s", PL, [I.IndexStr for I in IndictedList.Ilist])

    # Throw away new indictables that are already under active indictment.
    # Then build/append candidate spans to submit to rate analysis.
    # Incoming should always be left empty.  ENH: make Incoming a Queue
    @classmethod
    def FilterIncoming(cls, cbdata=None):
        """Event loop idle callback chained from AddIndictable()."""

        # FIXME: SQL every indictable before discard
        used = False
        while len(cls.Incoming):
            I = cls.Incoming.pop(0)
            lead = "Incoming %s %s" % (I.IndexStr, I.Diagonal())
            if I.physloc in cls.Indictments:
                logging.debug("%s: active PL indictment", lead)
                continue
            if I.physloc not in cls.Candidates:
                logging.debug("%s: new PL", lead)
                cls.Candidates[I.physloc] =  [ I ]
                used = True
            else:
                if I in cls.Candidates[I.physloc]:
                    logging.debug("%s: dup on PL", lead)
                    continue
                logging.debug("%s: add to PL", lead)
                cls.Candidates[I.physloc].append(I)
                used = True 
        if used:
            for v in cls.Candidates.values():
                v.sort(lambda a, b: a.TS - b.TS)
            REM.IdleAdd(cls.Candidate2Indicted)
        return False                # I am a callback: don't reschedule

    # Only add unique events, some analyzers may spew bursts of duplicates
    # (especially during testing).  The hash used by "in" must NOT include
    # the TS field.  The primary effect of this routine is to decouple the
    # event detection from an event loop "hot" routine to the idle loop.
    @classmethod
    def AddIndictable(cls, Inew):
        """Called by an analyzer when it detects an indictable event."""
        cls.IncomingTotal += 1
        if Inew in cls.Incoming:
            logging.debug("Duplicate event %s", Inew.IndexStr)
            return
        cls.IncomingUnique += 1
        cls.Incoming.append(Inew)   # "mostly-sorted" by increasing time
        if len(cls.Incoming) < 50:
            REM.IdleAdd(cls.FilterIncoming)
        else:
            logging.warning("'Incoming' queue is running away, size %d",
                len(cls.Incoming))
            cls.FilterIncoming()

    @classmethod
    def ExpireIndictments(cls, cbdata=None):
        """A periodic timer callback to remove aged events from queues."""

        TS = int(NOW())

        before = len(cls.Indictments)
        cls.Indictments = dict(
            [(k, v) for k, v in cls.Indictments.items() if v.alive(TS)])
        cls.IndictmentsExpired += before - len(cls.Indictments)

        # Rates are hourly, so 3-4 times an hour is fast enough
        TS = Globals.opts.verbose > 2 and 30 or 900
        REM.TimeoutAdd(TS, cls.ExpireIndictments, replace=True)
        return False    # Don't reschedule this callback instance

def ResetStats():
    """DBus callback, don't invoke from the program."""
    IndictQueues.IncomingTotal = 0
    IndictQueues.IncomingUnique = 0
    IndictQueues.IndictmentsExpired = 0

def Stats():
    """DBus callback, don't invoke from the program."""
    s = "IND: Incoming: %d (%d total, %d unique)\n" % (
        len(IndictQueues.Incoming),
        IndictQueues.IncomingTotal,
        IndictQueues.IncomingUnique)
    s += "IND: Candidates: %d\n" % len(IndictQueues.Candidates)
    s += "IND: Indictments: %d (%d expired)\n" % (
        len(IndictQueues.Indictments), IndictQueues.IndictmentsExpired)
    return s

def GetIndictments(throwaway):
    """DBus callback, don't invoke from the program."""
    logging.debug("GetIndictments")
    s = "There are %d indictments\n" % (len(IndictQueues.Indictments))
    for PL in IndictQueues.Indictments.keys():
        s += "%s\n" % repr(PL)
    return s

def ZapIndictments(throwaway, physlocRepr):
    """DBus callback, don't invoke from the program."""
    logging.debug("ZapIndictment(%s)", physlocRepr)
    return "There were %d indictments" % (len(IndictQueues.Indictments))
    IndictQueues.Indictments = { }

def Setup():
    """Called once from package __init__.py"""
    Globals.ResetStats.append(ResetStats)
    Globals.Stats.append(Stats)
    EmitSignal.ExposeMethod(GetIndictments, '', 's')
    EmitSignal.ExposeMethod(ZapIndictments, 's', 's')
    EmitSignal.ExposeSignal("IndictedPhysloc", "s")
    logging.info("Arm the expiration timeout")
    REM.IdleAdd(IndictQueues.ExpireIndictments)
