# HPDBus.py - no classes, just constants, variables, and support routines
"""Connect to DBus, establish the service name, and enable APIs."""

from __future__ import absolute_import	# Forces python 2.5

# import pdb
import logging
import sys
import time
# import inspect
# import new

import dbus
import dbus.service

from RASconfig import Globals

sys.modules[__name__].HP = True

# Constants

class HPDBus(object):
    """Hide the magic of DBus bindings for well-known operations.  The
    Gtk event loop must be established prior to invoking anything here."""

    # Constant values, "private" first

    _IPMIBusName = 'com.hp.m11y.IPMI'
    _RASBusName  = "com.hp.m11y.RAS"
    _IPMIobjbase = "/com/hp/m11y/IPMI"
    IRAS        = _RASBusName

    LogsInterface    = _IPMIBusName + '.Logs'
    SensorsInterface = _IPMIBusName + '.Sensors'

    SigAppended = "Appended"

    FPLobjpath = _IPMIobjbase + "/FPL"
    IMLobjpath = _IPMIobjbase + "/IML"
    SELobjpath = _IPMIobjbase + "/SEL"

    RASobjpath = "/com/hp/m11y/RAS"

    @classmethod
    def AddSigHandler(cls, sigpath, iface, signame, sighandler):
        """Wrapper for DBus 'add_signal_receiver()'"""

        # Runtime error: has this module been properly initialized?
        if not hasattr(HPDBus, "systembus"):
            return False

        if not callable(sighandler) or not len(sigpath) or not len(iface):
            raise LookupError   # Programmer error, not runtime
        HPDBus.systembus.add_signal_receiver(
            signal_name = signame,
            handler_function = sighandler,
            dbus_interface = iface,
            path_keyword = 'onpath',
            member_keyword = 'onmember')

        return True

    @classmethod
    def SEL_append_raw(cls, raw):
        """Send 14 raw bytes.  No idiot checking is done on this end."""
        try:
            cls._SELiface.Append(raw)
        except Exception, ex:
            logging.error("'SEL.Append' failed: %s", str(ex))
            return False
        return True

    @classmethod
    def GetBulk(cls, objbase, index, RecID, fwd, span):
        """Wrap the GetBulk method against hpipmid.  Returns a 3-ple:
           list of DBus events, nextIndex, nextRecID."""

        # Allow AttributeError
        ifaceobj = getattr(cls, "_" + objbase + "iface")

        try:    # DBus byte_arrays are a subclass of str, for binary data
            ret = ifaceobj.GetBulk(index, RecID, fwd, span, byte_arrays = True)
        except Exception:
            logging.exception("GetBulk() exception")
            return None, None, None

        if (len(ret) != 3):                 # list, nextIndex, nextRec
            logging.warning(
                "GetBulk(%d) returned %d args, expected 3", span, len(ret))
            raise TypeError
        DBusEvents, nextIndex, nextRecID = ret
        if not isinstance(DBusEvents, dbus.Array):
            logging.warning("Expected dbus.Array, got %s", type(DBusEvents))
            raise TypeError
        return DBusEvents, nextIndex, nextRecID

    @staticmethod
    def setup():
        """Connect to the system bus, establish the well-known name, and
        try to reach the IPMI caching daemon."""

        SYSTEMBUS = "systembus"
        if hasattr(HPDBus, SYSTEMBUS):  # singleton in single thread
            return False

        try:
            HPDBus.systembus = dbus.SystemBus()
        except Exception, ex:
            msg = "Could not connect to DBus System Bus: %s" % str(ex)
            logging.critical(msg)
            print >> sys.stderr, msg
            return False

        ret = HPDBus.systembus.request_name(
            HPDBus._RASBusName, dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
        if ret == dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER:
            logging.info("Got bus name %s", HPDBus._RASBusName)
        else:
            if ret == dbus.bus.REQUEST_NAME_REPLY_EXISTS:
                msg = "Another process already owns %s" % HPDBus._RASBusName
            elif ret == dbus.bus.REQUEST_NAME_REPLY_ALREADY_OWNER:
                msg = "I already own %s" % HPDBus._RASBusName
            else:
                msg = "Can't get DBus name %s and don't know why" % \
                    HPDBus._RASBusName
            logging.critical(msg)
            print >> sys.stderr, msg
            return False

        # Establish objects that connect to IPMI caching daemon.
        # ENH: this needs to be more retry-based for long-term reliability
        # Maybe loop on this, instead of try/except?
        if not HPDBus._IPMIBusName in map(str, HPDBus.systembus.list_names()):
            logging.warning(
                "%s is not present on System DBus", HPDBus._IPMIBusName)

        timeout = 10
        while True:
            try:
                # GTK mainloop must be established before this call
                HPDBus._SELproxy = HPDBus.systembus.get_object(
                    HPDBus._IPMIBusName, HPDBus.SELobjpath)
                HPDBus._SELiface = dbus.Interface(
                    HPDBus._SELproxy, dbus_interface = HPDBus.LogsInterface)
                break
            except Exception, ex:   # get_object
                if timeout > 0:
                    logging.warning("retry, %d secs remaining", timeout)
                    time.sleep(2)
                    timeout -= 2
                    continue
                if isinstance(ex, dbus.DBusException) and \
                        "was not provided" in str(ex):
                    logging.critical(
                    "No IPMI log daemon present on DBus")
                else:
                    logging.critical(
                        "Cannot connect to DBus IPMI daemon: %s", str(ex))
                delattr(HPDBus, SYSTEMBUS)
                return False                 # ENH: retry via REM timer?

        try:
            HPDBus._IMLproxy = HPDBus.systembus.get_object(
                HPDBus._IPMIBusName, HPDBus.IMLobjpath)
            HPDBus._IMLiface = dbus.Interface(
                HPDBus._IMLproxy, dbus_interface = HPDBus.LogsInterface)
        except dbus.DBusException:
            HPDBus._IMLiface = None
            HPDBus._IMLproxy = None

        try:
            HPDBus._FPLproxy = HPDBus.systembus.get_object(
                HPDBus._IPMIBusName, HPDBus.FPLobjpath)
            HPDBus._FPLiface = dbus.Interface(
                HPDBus._FPLproxy, dbus_interface = HPDBus.LogsInterface)
        except dbus.DBusException:
            HPDBus._FPLiface = None
            HPDBus._FPLproxy = None

        return True

###########################################################################

# Subclassing dbus.service.Object and calling its __init__ will export the
# path and interface with decorated methods and signals defined in the class.
# You also get an org.freedesktop.DBus.Introspectable interface with an
# Introspect() that has all the right stuff.  Very cool.
# ENH: make this a factory for arbitrary interfaces and signals

class EmitSignal(dbus.service.Object):
    """Groundwork for emitting a DBus signal.  This is a singleton to
    avoid multiple binding errors from dbus.  Some of the internal
    comments about SUPPORTS_MULTIPLE_[CONNECTIONS|OBJECTS] made me
    a little nervous."""

    _theOne = None   # scalar makes pychecker whine when re-assigned

    # By magic, this is a static method that is still passed a class.
    # It's not quite ready for prime time, as the "dbus.server.Object"
    # call in __init__ doesn't work when put here.
    # FIXME: verify that last statement

    def __new__(cls, *args, **kargs):
        if not cls._theOne:
            cls._theOne = super(EmitSignal, cls).__new__(
                cls, *args, **kargs)
        return cls._theOne

    def __init__(self):
        """Call the superclass and set up a singleton.  The Introspection
        can't be revealed until the event loop is running and can respond
        to the method call."""
        if not hasattr(self._theOne, "HPDBus"):
            self._theOne.HPDBus = True
            dbus.service.Object.__init__(
                self,
                conn = HPDBus.systembus,
                object_path = HPDBus.RASobjpath,
                bus_name = None)    # FIXME: can this be :1.xxxxx/com.hp.IPMI?

    # Decorator assumes it's a class/instance method of something and
    # immediately pops and tosses the first arg cuz it doesn't care,
    # duck-typing for the fields it wants.  Either pass in a class or
    # instance method, or dfine it with a 1st argument that's unused.
    @classmethod
    def ExposeMethod(cls, origfunc, insig, outsig):
        """The first argument of your method will be tossed."""
        if not callable(origfunc):
            raise TypeError("origfunc is not a callable")
        methname = origfunc.__name__
        if hasattr(cls, methname):
            raise KeyError(methname + "already exposed")
        decorated = dbus.service.method(
                dbus_interface=HPDBus.IRAS,
                in_signature=insig,
                out_signature=outsig)(origfunc)
        setattr(cls, methname, decorated)
        # inspired by python2.6/site/dbus/service.py::Introspect()
        FQCN = "%s.%s" % (cls.__module__, cls.__name__)
        interfaces = cls._dbus_class_table[FQCN]
        if not interfaces.has_key(HPDBus.IRAS):
            interfaces[HPDBus.IRAS] = { }
        interfaces[HPDBus.IRAS][methname] = origfunc

    @classmethod
    def dispatch(cls, signame, arg1):
        """Called when you want to emit a DBus signal.  ExposeSignal() must
        be called prior to this to setup the signal by name."""
        logging.debug("Emitting DBus signal '%s'", signame)
        meth = getattr(cls, signame, None)
        if not meth:
            logging.error("No such signal %s", signame)
            return
        meth(cls._theOne, arg1)

    # Never call templates directly.  At DBus introspection time the arg
    # count and names are extracted from # the code object
    #(cls._sigtemplate.func_code.co_args and co_varnames) which are read only.
    # Thus I can't easily change the name or count at runtime.  The
    # signame passed to ExposeSignal will have to be enough context.
    # DBus library methods are written with instance methods (inherited
    # from dbus.service.Object) in mind.  The first thing they do is
    # toss "self", but I still have to account for it.
    @staticmethod
    def _sigtemplate(throwaway, arg1):
        """Never call this."""
        pass

    @classmethod
    def ExposeSignal(cls, signame, sigsignature):
        """For now, signature must be a single simple type."""
        if hasattr(EmitSignal, signame):
            raise KeyError(signame + "already exposed")
        if 'a' in sigsignature or \
           'v' in sigsignature or \
           '(' in sigsignature:
            raise ValueError("signature is not simple type")
        if len(sigsignature) != 1:
            raise ValueError("signature is wrong length")
        # Saved by the decorator maker (via closure) as "member_name"
        cls._sigtemplate.__name__ = signame
        decorated = dbus.service.signal(
                dbus_interface=HPDBus.IRAS,
                signature=sigsignature)(cls._sigtemplate)
        decorated.__name__ = decorated.func_name = signame
        setattr(EmitSignal, signame, decorated)
        FQCN = "%s.%s" % (cls.__module__, cls.__name__)
        interfaces = EmitSignal._dbus_class_table[FQCN]
        interfaces[HPDBus.IRAS][signame] = decorated

def setup():
    """Defer to the classes.  Ignore the instances, as they are saved as
    singletons within the class."""

    if not HPDBus.setup():
        return False
    EmitSignal()

    # Don't call any of the following functions directly.  They are
    # callbacks for DBus method invocations of the same name.  The
    # DBus bindings expect to see an instance method, meaning they'll
    # be expecting a "self" argument.  They never use it.

    def GetStats(throwaway):
        """Invoke this method from DBus to see internal statistics."""
        logging.debug("Get statistics")
        allstats = ""
        for mod in Globals.Stats:
            allstats += mod()
        return allstats
    EmitSignal.ExposeMethod(GetStats, '', 's')

    def ResetStats(throwaway):
        """Invoke this method from DBus to clear internal statistics."""
        logging.debug("Reset statistics")
        for mod in Globals.ResetStats:
            mod()
    EmitSignal.ExposeMethod(ResetStats, '', '')

    # Export test members in debug mode.  Invoke the method to
    # generate the signals.
    if Globals.opts.verbose > 2:
        def testMethod(throwaway, mystr, myint):
            """Call this from DBus to emit two test signals."""
            fullstr = "'%s' %d" % (mystr, myint)
            logging.debug("DBus test method: %s", fullstr)
            EmitSignal.dispatch("testSigInt", myint)
            EmitSignal.dispatch("testSigStr", mystr)
            return "You sent " + fullstr
        EmitSignal.ExposeMethod(testMethod, 'si', 's')
        EmitSignal.ExposeSignal("testSigInt", 'i')
        EmitSignal.ExposeSignal("testSigStr", 's')

    return True
