# HPphysloc - a class for the HP SPPA Physical Location spec
"""SPPA 3.0 Physical Location"""

import sys

sys.modules[__name__].HP = True

class _Fields(object):
    """Descriptor to calculate fields from raw bytes."""

    def __init__(self, offset):
        if not isinstance(offset, int):
            raise AttributeError("offset is not an integer")
        if offset < -2 or offset == 0 or offset > 7:
            raise AttributeError("offset is out of range")
        self.offset = offset

    def __get__(self, instance, unused_owner=None):
        if self.offset > 0 and self.offset < 8:
            return ord(instance.raw[self.offset])
        if self.offset == -2:   # detail
            return ord(instance.raw[0]) & 0x0F
        if instance.offset == -1:   # detail
            return (ord(instance.raw[0]) >> 4) & 0x0F
        raise KeyError("phyloc field descriptor is corrupt")

    def __set__(self, instance, value):
        raise NotImplementedError

class Physloc(object):
    """SPPA EAS 3.0 Physical Location is 8 bytes of 'level' data
       (enclosure hierarchy) and a description string."""

    __slots__ = [ "detail", "source", "subslot", "slot", "ext",
                  "cell", "ioc", "bay", "cab", "raw", "desc" ]

    detail  = _Fields(-2)
    source  = _Fields(-1)
    subslot = _Fields(1)
    slot    = _Fields(2)
    ext     = _Fields(3)
    cell    = _Fields(4)
    ioc     = _Fields(5)
    bay     = _Fields(6)
    cab     = _Fields(7)

    def __init__(self, bytestr = None, desc = None):
        """bytestr is usually a little-endian packed byte array (taken
        from a DBus Array).  It could also be a string of the form
        AA-BB-CC-DD-EE-FF-GG-HH  (AA == cabinet).  If bytestr is missing,
        set it to the NULL physloc FF-FF-FF-FF-FF-FF-FF-00."""

        if bytestr is None:
            self.setNULL()
            return
        if not isinstance(bytestr, str):
            raise AttributeError("bytestr must be a string")
        self.desc = desc and desc
        # Store raw as a string (ie, native byte array) because that makes
        # for the fastest methods below.
        if len(bytestr) == 8:
            self.raw = bytestr
        elif len(bytestr) == 23 and bytestr.count('-') == 7:
            bytelist = bytestr.split('-')
            self.raw = "".join(chr(int(h2, base=16)) for h2 in bytelist[::-1])
        else:
            raise AttributeError("bytestr format error")

    def __repr__(self):
        """This only returns XX-XX-XX-XX-XX-XX-XX-XX and drops desc."""
        return "-".join(["%02X" % ord(c) for c in self.raw[::-1]])

    def __str__(self):
        """A simple-minded substitute for libezbmc 'BMC_physloc_parse()'."""
        s = 'Python: '
        if self.cab     != 0xFF:
            s = s + "cab=%d "     % self.cab
        if self.bay     != 0xFF:
            s = s + "bay=%d "     % self.bay
        if self.ioc     != 0xFF:
            s = s + "ioc=%d "     % self.ioc
        if self.cell    != 0xFF:
            s = s + "cell=%d "    % self.cell
        if self.ext     != 0xFF:
            s = s + "ext=%d "     % self.ext
        if self.slot    != 0xFF:
            s = s + "slot=%d "    % self.slot
        if self.subslot != 0xFF:
            s = s + "subslot=%d " % self.subslot
        if len(s) > 0 or self.source != 0:
            s = s + "src=%d " % self.source
        if len(s) > 0 or self.detail != 0:
            s = s + "detail=%d" % self.detail
        s = s + "\nezbmc:  %s" % ( self.desc )
        return s

    def __hash__(self):     # at 2.5 can use long int
        """self.id is misleading with multiple objects, so hash on raw."""
        return hash(self.raw)

    def __cmp__(self, other):
        """Not clear if this will ever be invoked since I need __eq__ and
        __hash__."""
        raise NotImplementedError

    def __eq__(self, other):
        """Assumes the raw field is the basis for all others."""
        return self.raw == other.raw

    def isNULL(self):
        """Test for equality to the NULL physloc"""
        return self.raw == "\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF"

    def setNULL(self):
        """Set an instance to the NULL physloc."""
        self.raw = "\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        self.desc = "NULL physloc"

if __name__ == "__main__":

    Pnull = Physloc()
    print "create NULL physloc:", repr(Pnull) == "FF-FF-FF-FF-FF-FF-FF-00"

    pLE = Physloc("\0\1\2\3\4\5\6\7",        "Little Endian")
    pBE = Physloc("07-06-05-04-03-02-01-00", "Big Endian")
    print "Different endian constructors equiv:", pLE == pBE

    try:
        Pnull.cab = 5
        result = False
    except NotImplementedError:
        result = True
    print "Reject changing an attribute:", result

    try:
        Pnull.badboy = 5
        result = False
    except AttributeError:
        result = True
    print "Reject adding new attributes:", result

    p1   = Physloc("\1\1\1\1\1\1\1\1", "one")
    pone = Physloc("\1\1\1\1\1\1\1\1", "one")
    d = { }
    d[p1] = "This is p1"
    print "Container test 1:", p1 in d         # Doesn't call __eq__
    print "Container test 2:", pone in d       # DOES call eq
